#pragma once

#include <list>
#include <vector>

#include <cppcore/conversion/string.h>
#include <cppcore/conversion/convert_cast.h>
#include <cppargs/misc/option_parser.h>

namespace cppargs
{

    template<typename T_derived, typename T_value>
    struct option_parser_list_impl;

    template<typename T_value>
    void option_parser<T_value>::operator()() const
    {
        auto arg = context.arg;
        auto str = context.value;

        if (!str)
        {
            auto ret = context.next();
            if (ret != next_result::argument)
                throw std::runtime_error(std::string("expected one argument (option=") + arg + ")");
            str = context.arg;
        }

        if (!cppcore::try_from_string(str, value))
            throw std::runtime_error(std::string("unable to parse option argument (option=") + arg + "; value=" + str + ")");
    }

    template<>
    struct option_parser<bool>
    {
        cppargs::context&   context;
        bool&               value;

        inline void operator()() const
        {
            value = true;
            auto arg = context.arg;
            auto str = context.value;

            if (!str)
            {
                auto next_context = context;
                auto ret = next_context.next();
                if (ret == next_result::argument)
                {
                    context = next_context;
                    str = context.arg;
                }
            }

            if (str && !cppcore::try_from_string(str, value))
                throw std::runtime_error(std::string("unable to parse option argument (option=") + arg + "; value=" + str + ")");
        }
    };

    template<typename T_value>
    struct option_parser<std::vector<T_value>>
        : public option_parser_list_impl<
            option_parser<std::vector<T_value>>, T_value>
    {
        using value_type  = T_value;
        using vector_type = std::vector<value_type>;
        using this_type   = option_parser<vector_type>;
        using base_type   = option_parser_list_impl<this_type, value_type>;

        cppargs::context&   context;
        vector_type&        value;

        inline option_parser(
            cppargs::context& p_context,
            vector_type&      p_value)
            : base_type (*this)
            , context   (p_context)
            , value     (p_value)
            { }

        template<typename X>
        inline void emplace(X&& x)
            { value.emplace_back(std::forward<X>(x)); }
    };

    template<typename T_value>
    struct option_parser<std::list<T_value>>
        : public option_parser_list_impl<
            option_parser<std::list<T_value>>, T_value>
    {
        using value_type  = T_value;
        using list_type   = std::list<value_type>;
        using this_type   = option_parser<list_type>;
        using base_type   = option_parser_list_impl<this_type, value_type>;

        cppargs::context&   context;
        list_type&          value;

        inline option_parser(
            cppargs::context& p_context,
            list_type&      p_value)
            : base_type (*this)
            , context   (p_context)
            , value     (p_value)
            { }

        template<typename X>
        inline void emplace(X&& x)
            { value.emplace_back(std::forward<X>(x)); }
    };

    template<typename T_derived, typename T_value>
    struct option_parser_list_impl
    {
        using derived_type = T_derived;
        using value_type   = T_value;

        derived_type& derived;

        inline option_parser_list_impl(
            derived_type& p_derived)
            : derived(p_derived)
            { }

        inline void operator()() const
        {
            auto arg = derived.context.arg;
            auto index = derived.context.index;
            if (derived.context.value)
            {
                const char * s = derived.context.value;
                const char * e = s - 1;
                do
                {
                    ++e;
                    if (    s != e
                        && (    *e == ','
                            ||  *e == '\0'))
                    {
                        value_type tmp;
                        std::string str(s, cppcore::convert_cast<size_t>(e - s));
                        s = e + 1;
                        if (!cppcore::try_from_string(str, tmp))
                            throw std::runtime_error(std::string("unable to parse option argument (option=") + arg + "; value=" + s + ")");
                        derived.emplace(std::move(tmp));
                    }
                }
                while(*e != '\0');
            }
            else
            {
                auto new_context = derived.context;
                while (new_context.next() == next_result::argument)
                {
                    value_type tmp;
                    if (!cppcore::try_from_string(new_context.arg, tmp))
                        throw std::runtime_error(std::string("unable to parse option argument (option=") + arg + "; value=" + new_context.arg + ")");
                    derived.emplace(std::move(tmp));
                    derived.context = new_context;
                }
                if (derived.context.index == index)
                    throw std::runtime_error(std::string("expected at least one argument (option=") + arg + ")");
            }
        }
    };

}
