#pragma once

#include <string>
#include <cppargs/misc/context.h>

namespace cppargs
{

    /**
     * @brief Class to store all meta data for a option.
     */
    struct option_meta
    {
    public:
        char        short_name      { '\0'  };  //!< short option of the option (without the prependding -)
        std::string long_name       { ""    };  //!< long name of the option (without the prepending --)
        std::string description     { ""    };  //!< description for this option
        std::string arguments       { ""    };  //!< arguments to pass to this option
        std::string default_value   { ""    };  //!< default value for this option
        bool        mandatory       { false };  //!< TRUE if this option is optional, FALSE otherwise

    public:
        /**
         * @brief Set the sort name of the option.
         */
        inline option_meta& set_short_name(char value);

        /**
         * @brief Set the ling name of the option.
         */
        inline option_meta& set_long_name(const std::string& value);

        /**
         * @brief Set description of the option.
         */
        inline option_meta& set_description(const std::string& value);

        /**
         * @brief Set the argument description.
         */
        inline option_meta& set_arguments(const std::string& value);

        /**
         * @brief Set the default value of the option.
         */
        template <typename T_value>
        inline option_meta& set_default_value(const T_value& value);

        /**
         * @brief Set to true if the option is mandatorey, false otherwise.
         */
        inline option_meta& set_mandatory(bool value = true);

    public:
        /**
         * @brief Set the default values for the option depending on the passed value.
         *
         * @param[in] meta  Meta object to set default values for.
         * @param[in] value Value to get default values from.
         *
         * @return          New meta object with the default values for the passed value.
         */
        template<typename T_value>
        static option_meta prepare(const option_meta& meta, T_value&& value);
    };

    /**
     * @brief Option to parse from command line arguments
     */
    struct option
    {
    public:
        const option_meta meta; //!< meta data of the option

    protected:
        /**
         * @brief Constructor.
         *
         * @param[in] meta  Meta data for this option.
         */
        inline option(const option_meta& meta);

    public:
        /**
         * @brief Destructor.
         */
        virtual ~option() = default;

        /**
         * @brief Parse the option using the current context.
         *
         * @param[in,out] c     Context to use for parsing.
         */
        virtual void parse(context& c) const = 0;
    };

}
