#pragma once

#include <cppargs/parser/parser.h>
#include <cppargs/group/simple_group.h>

namespace cppargs
{

    /**
     * @brief Class to parse options using simple types or callbacks.
     */
    struct simple_parser
        : public parser
    {
    public:
        using base_type       = parser;
        using option_ptr_type = typename base_type::option_ptr_type;

    private:
        simple_group _group;

    public:
        /**
         * @brief Constructor.
         *
         * @param[in] p_default     Default option to use for unknown arguents (optional / can be null).
         * @param[in] p_args        Options and subgroups.
         */
        template<typename... T_args>
        simple_parser(
            option_ptr_type&& p_default,
            T_args&&... p_args);

        /**
         * @brief Parse the command line arguments using the passed options.
         *
         * @param[in] argc  Number of arguments in argv.
         * @param[in] argv  Array of all arguments.
         */
        inline void parse(int argc, char const * argv[]) const;

    public:
        /**
         * @brief Create new option group.
         *
         * @param[in] meta      Meta data of the new group.
         * @param[in] args      Subgroups and options.
         *
         * @return              Created group.
         */
        template<typename... T_args>
        static constexpr decltype(auto) make_group(const group_meta& meta, T_args&&... args);

        /**
         * @brief Create new predicate option.
         *
         * @param[in] meta      Meta data of the new option.
         * @param[in] predicate Predicate to execute for this option.
         *
         * @return              Created option.
         */
        template<typename T_predicate>
        static constexpr decltype(auto) make_predicate_option(const option_meta& meta, T_predicate&& predicate);

        /**
         * @brief Create new reference option.
         *
         * @param[in] meta      Meta data of the new option.
         * @param[in] value     Value to set for this option.
         *
         * @return              Created option.
         */
        template<typename T_value>
        static constexpr decltype(auto) make_reference_option(const option_meta& meta, T_value& value);
    };

}
